# -----------------------------------------------------------------------------
# $Id: target.mk,v 1.4 1999/12/07 16:04:17 simonmar Exp $
#
# hslibs/target.mk
#

# We need to set TOP to be the TOP that the next level up expects!
# The TOP variable is reset after the inclusion of the fptools
# boilerplate, so we stash TOP away first:
HSLIBS_TOP := $(TOP)
TOP:=$(TOP)/..

# -----------------------------------------------------------------------------
# Makefiles in the subdirectories should set the following variables:
# 
#   HSLIB		the (short) name of the top-level library category 
#			we're building.  eg. lang, net, text etc.
#
#   HSLIB_DEPS  	other library categories that this category depends on.
#
#   SRCS_FROM_SUBDIRS	pull out *.hs and *.lhs sources from these subdirectories
#			under the current directory.
#
#   NOT_FOR_GHC		exclude source files from certain compilers.
#   NOT_FOR_NHC
#   NOT_FOR_HBC
#   NOT_FOR_HUGS

# -----------------------------------------------------------------------------
# Filter source file list depending on which compiler we're using

ifeq "$(HsLibsFor)" "ghc"
SRCS := $(filter-out $(NOT_FOR_GHC), $(SRCS))
HC = $(FPTOOLS_TOP)/ghc/driver/ghc-inplace
MKDEPENDHS = $(HC)
endif

ifeq "$(HsLibsFor)" "hugs"
SRCS := $(filter-out $(NOT_FOR_HUGS), $(SRCS))
MKDEPENDHS = $(GHC)
endif

ifeq "$(HsLibsFor)" "nhc"
SRCS := $(filter-out $(NOT_FOR_NHC), $(SRCS))
HC = $(NHC)
MKDEPENDHS = $(GHC)
endif

ifeq "$(HsLibsFor)" "hbc"
SRCS := $(filter-out $(NOT_FOR_HBC), $(SRCS))
HC = $(HBC)
MKDEPENDHS = $(GHC)
endif

# -----------------------------------------------------------------------------
# Building libraries for GHC

ifeq "$(HsLibsFor)" "ghc"

ifneq "$(HSLIB)" ""

LIBRARY = libHS$(HSLIB)$(_way).a
LIBOBJS = $(HS_OBJS)

WAYS=$(GhcLibWays)

SRCS += $(wildcard $(patsubst %, %/*hs, $(SRCS_FROM_SUBDIRS)))
SRC_HC_OPTS += $(patsubst %, -i%, $(SRCS_FROM_SUBDIRS))

MAGIC_HSCPP_OPTS=-DBEGIN_FOR_GHC='-}' -DEND_FOR_GHC='{-' -DBEGIN_FOR_HUGS='{-' -DEND_FOR_HUGS='-}'
SRC_MKDEPENDHS_OPTS += -optdep-w $(MAGIC_HSCPP_OPTS)

ifneq "$(way)" "dll"
SRC_HC_OPTS += -static
endif

#
# Object and interface files have suffixes tagged with their ways
#
ifneq "$(way)" ""
SRC_HC_OPTS += -hisuf $(way_)hi
endif

SRC_HC_OPTS += $(GhcLibHcOpts)

DLL_NAME = HS$(HSLIB).dll
DLL_IMPLIB_NAME = libHS$(HSLIB)_imp.a
SRC_BLD_DLL_OPTS += --export-all --output-def=HS$(HSLIB).def DllVersionInfo.o
SRC_BLD_DLL_OPTS += -lwinmm -lHSrts_imp -lHScbits_imp -lHS_imp -lgmp -L. -L$(GHC_RTS_DIR)/gmp -L$(GHC_RTS_DIR) -L$(GHC_LIB_DIR)/std -L$(GHC_LIB_DIR)/std/cbits

ifeq "$(way)" "dll"
all :: DllVersionInfo.o
endif

# add syslib depenencies
SRC_HC_OPTS += $(patsubst %, -syslib %, $(HSLIB_DEPS))

# Installation; need to install .hi files as well as libraries
#
# The interface files are put inside the $(libdir), since they
# might (potentially) be platform specific..
#
# override is used here because for binary distributions, datadir is
# set on the command line. sigh.
#

override datadir:=$(libdir)/imports/$(HSLIB)
INSTALL_LIBS  += $(LIBRARY)
INSTALL_DATAS += $(HS_IFACES)

ifeq "$(EnableWin32DLLs)" "YES"
INSTALL_PROGS += $(DLL_NAME)
INSTALL_LIBS  += $(patsubst %.a, %_imp.a, $(LIBRARY)) 
INSTALL_DATAS += dLL_ifs.hi
endif

endif # HSLIB /= ""

endif # HsLibsFor = ghc

# -----------------------------------------------------------------------------
# Building libraries for Hugs

# we CPP all the haskell sources and pop them in a mirror hierarchy
# under hugs-srcs/.

ifeq "$(HsLibsFor)" "hugs"

HUGS_SRC_DIR = hugs-srcs
HUGS_SRC_DIRS = $(HUGS_SRC_DIR) $(patsubst %, $(HUGS_SRC_DIR)/%, $(SRCS_FROM_SUBDIRS))

SRCS += $(wildcard $(patsubst %, %/*hs, $(SRCS_FROM_SUBDIRS)))
HUGS_SRCS = $(patsubst %, $(HUGS_SRC_DIR)/%, $(SRCS))

all :: $(HUGS_SRC_DIRS) $(HUGS_SRCS)

$(HUGS_SRC_DIRS) ::
	$(MKDIRHIER) $@

HUGS_DEFINES = -D__HUGS__

$(HUGS_SRC_DIR)/%.hs : %.hs
	$(RAWCPP) $(HUGS_DEFINES) -o $(HUGS_SRC_DIR)/$*.hs - < $<

$(HUGS_SRC_DIR)/%.lhs : %.lhs
	$(RAWCPP) $(HUGS_DEFINES) -o $(HUGS_SRC_DIR)/$*.lhs - < $<

endif

# -----------------------------------------------------------------------------

include $(TOP)/mk/target.mk

# Reset TOP
TOP:=$(HSLIBS_TOP)
